@@ -311,7 +311,7 @@ got_get_repo_commit(struct request *c, struct repo_commit *repo_commit,
 }
 
 const struct got_error *
-got_get_repo_commits(struct request *c, int limit)
+got_get_repo_commits(struct request *c, size_t limit)
 {
 	const struct got_error *error = NULL;
 	struct got_object_id *id = NULL;
@@ -326,11 +326,23 @@ got_get_repo_commits(struct request *c, int limit)
 	struct querystring *qs = t->qs;
 	struct repo_dir *repo_dir = t->repo_dir;
 	char *in_repo_path = NULL, *repo_path = NULL, *file_path = NULL;
-	int chk_next = 0, chk_multi = 0;
+	int chk_next = 0;
+
+	if (limit == 0)
+		return got_error(GOT_ERR_RANGE);
+
+	if (limit > 1) {
+		/*
+		 * Traverse one commit more than requested to provide
+		 * the next button.
+		 */
+		limit++;
+		chk_next = 1;
+	}
 
 	TAILQ_INIT(&refs);
 
-	if (qs->file != NULL && strlen(qs->file) > 0)
+	if (qs->file != NULL && *qs->file != '\0')
 		if (asprintf(&file_path, "%s/%s", qs->folder ? qs->folder : "",
 		    qs->file) == -1)
 			return got_error_from_errno("asprintf");
@@ -364,7 +376,7 @@ got_get_repo_commits(struct request *c, int limit)
 	if (error)
 		goto done;
 
-	if (qs->file != NULL && strlen(qs->file) > 0) {
+	if (qs->file != NULL && *qs->file != '\0') {
 		error = got_commit_graph_open(&graph, file_path, 0);
 		if (error)
 			goto done;
@@ -404,38 +416,18 @@ got_get_repo_commits(struct request *c, int limit)
 			goto done;
 		}
 
+		if (--limit == 0 && chk_next) {
+			t->more_id = strdup(repo_commit->commit_id);
+			if (t->more_id == NULL)
+				error = got_error_from_errno("strdup");
+			goto done;
+		}
+
 		TAILQ_INSERT_TAIL(&t->repo_commits, repo_commit, entry);
 
-		if (!chk_multi || limit != 1 ||
-		    srv->max_commits_display == 1) {
-			chk_multi = 1;
+		if (limit == 0)
+			goto done;
 
-			/*
-			 * check for one more commit before breaking,
-			 * so we know whether to navigate through briefs
-			 * commits and summary
-			 */
-			if (chk_next && (qs->action == BRIEFS ||
-			    qs->action == COMMITS || qs->action == SUMMARY)) {
-				t->more_id = strdup(repo_commit->commit_id);
-				if (t->more_id == NULL) {
-					error = got_error_from_errno("strdup");
-					goto done;
-				}
-				got_object_commit_close(commit);
-				commit = NULL;
-				TAILQ_REMOVE(&t->repo_commits, repo_commit,
-				    entry);
-				gotweb_free_repo_commit(repo_commit);
-				goto done;
-			}
-		}
-		if (error || (limit && --limit == 0)) {
-			if (qs->file != NULL && *qs->file != '\0')
-				if (chk_multi == 0)
-					break;
-			chk_next = 1;
-		}
 		if (commit) {
 			got_object_commit_close(commit);
 			commit = NULL;
@@ -457,7 +449,7 @@ got_get_repo_commits(struct request *c, int limit)
 }
 
 const struct got_error *
-got_get_repo_tags(struct request *c, int limit)
+got_get_repo_tags(struct request *c, size_t limit)
 {
 	const struct got_error *error = NULL;
 	struct got_object_id *id = NULL;
@@ -479,6 +471,9 @@ got_get_repo_tags(struct request *c, int limit)
 
 	TAILQ_INIT(&refs);
 
+	if (limit == 0)
+		return got_error(GOT_ERR_RANGE);
+
 	if (asprintf(&repo_path, "%s/%s", srv->repos_path,
 	    repo_dir->name) == -1)
 		return got_error_from_errno("asprintf");
