@@ -131,7 +131,7 @@
 #define GOT_ERR_GITCONFIG_SYNTAX 114
 #define GOT_ERR_REBASE_OUT_OF_DATE 115
 #define GOT_ERR_CACHE_DUP_ENTRY	116
-#define GOT_ERR_QUERYSTRING	117
+/* 117 is currently unused */
 #define GOT_ERR_FETCH_FAILED	118
 #define GOT_ERR_PARSE_URI	119
 #define GOT_ERR_BAD_PROTO	120
@@ -171,7 +171,7 @@
 #define GOT_ERR_BAD_TAG_SIGNATURE 154
 #define GOT_ERR_VERIFY_TAG_SIGNATURE 155
 #define GOT_ERR_SIGNING_TAG	156
-/* 157 is currently unused */
+#define GOT_ERR_BAD_OPTION 157
 #define GOT_ERR_BAD_QUERYSTRING	158
 #define GOT_ERR_INTEGRATE_BRANCH 159
 #define GOT_ERR_BAD_REQUEST	160
