@@ -610,10 +610,46 @@ done:
 	return err;
 }
 
+const struct got_error *
+got_repo_cleanup_prepare(struct got_repository *repo,
+    struct got_lockfile **lk)
+{
+	const struct got_error *err;
+	char myname[_POSIX_HOST_NAME_MAX + 1];
+
+	if (gethostname(myname, sizeof(myname)) == -1)
+		return got_error_from_errno("gethostname");
+
+	err = got_lockfile_lock(lk, "gc.pid", got_repo_get_fd(repo));
+	if (err)
+		return err;
+
+	/*
+	 * Git uses these info to provide some verbiage when finds a
+	 * lock during `git gc --force' so don't try too hard to avoid
+	 * short writes and don't care if a race happens between the
+	 * lockfile creation and the write itself.
+	 */
+	if (dprintf((*lk)->fd, "%d %s", getpid(), myname) < 0)
+		return got_error_from_errno("dprintf");
+
+	return NULL;
+}
+
+const struct got_error *
+got_repo_cleanup_complete(struct got_repository *repo,
+    struct got_lockfile *lk)
+{
+	if (lk == NULL)
+		return NULL;
+
+	return got_lockfile_unlock(lk, got_repo_get_fd(repo));
+}
+
 static const struct got_error *
 report_cleanup_progress(got_cleanup_progress_cb progress_cb,
     void *progress_arg, struct got_ratelimit *rl,
-    int nloose, int ncommits, int npurged)
+    int nloose, int ncommits, int npurged, int nredundant)
 {
 	const struct got_error *err;
 	int elapsed;
@@ -625,7 +661,7 @@ report_cleanup_progress(got_cleanup_progress_cb progress_cb,
 	if (err || !elapsed)
 		return err;
 
-	return progress_cb(progress_arg, nloose, ncommits, npurged);
+	return progress_cb(progress_arg, nloose, ncommits, npurged, nredundant);
 }
 
 static const struct got_error *
@@ -720,7 +756,8 @@ get_loose_object_ids(struct got_object_idset **loose_ids, off_t *ondisk_size,
 				goto done;
 			err = report_cleanup_progress(progress_cb,
 			    progress_arg, rl,
-			    got_object_idset_num_elements(*loose_ids), -1, -1);
+			    got_object_idset_num_elements(*loose_ids),
+			    -1, -1, -1);
 			if (err)
 				goto done;
 		}
@@ -1016,7 +1053,7 @@ load_commit_or_tag(struct got_object_idset *loose_ids, int *ncommits,
 			(*ncommits)++; /* scanned tags are counted as commits */
 
 		err = report_cleanup_progress(progress_cb, progress_arg, rl,
-		    nloose, *ncommits, -1);
+		    nloose, *ncommits, -1, -1);
 		if (err)
 			break;
 
@@ -1105,7 +1142,7 @@ purge_loose_object(struct got_object_id *id, void *data, void *arg)
 		a->npurged++;
 		a->size_purged += sb.st_size;
 		err = report_cleanup_progress(a->progress_cb, a->progress_arg,
-		    a->rl, a->nloose, a->ncommits, a->npurged);
+		    a->rl, a->nloose, a->ncommits, a->npurged, -1);
 		if (err)
 			goto done;
 	}
@@ -1120,15 +1157,16 @@ done:
 
 const struct got_error *
 got_repo_purge_unreferenced_loose_objects(struct got_repository *repo,
-    off_t *size_before, off_t *size_after, int *npacked, int dry_run,
-    int ignore_mtime, got_cleanup_progress_cb progress_cb, void *progress_arg,
+    off_t *size_before, off_t *size_after, int *ncommits, int *nloose,
+    int *npacked, int dry_run, int ignore_mtime,
+    got_cleanup_progress_cb progress_cb, void *progress_arg,
     got_cancel_cb cancel_cb, void *cancel_arg)
 {
 	const struct got_error *err;
 	struct got_object_idset *loose_ids;
 	struct got_object_idset *traversed_ids;
 	struct got_object_id **referenced_ids;
-	int i, nreferenced, nloose, ncommits = 0;
+	int i, nreferenced;
 	struct got_reflist_head refs;
 	struct got_reflist_entry *re;
 	struct purge_loose_object_arg arg;
@@ -1146,11 +1184,11 @@ got_repo_purge_unreferenced_loose_objects(struct got_repository *repo,
 	    progress_cb, progress_arg, &rl, repo);
 	if (err)
 		return err;
-	nloose = got_object_idset_num_elements(loose_ids);
-	if (nloose == 0) {
+	*nloose = got_object_idset_num_elements(loose_ids);
+	if (*nloose == 0) {
 		got_object_idset_free(loose_ids);
 		if (progress_cb) {
-			err = progress_cb(progress_arg, 0, 0, 0);
+			err = progress_cb(progress_arg, 0, 0, 0, -1);
 			if (err)
 				return err;
 		}
@@ -1188,9 +1226,9 @@ got_repo_purge_unreferenced_loose_objects(struct got_repository *repo,
 
 	for (i = 0; i < nreferenced; i++) {
 		struct got_object_id *id = referenced_ids[i];
-		err = load_commit_or_tag(loose_ids, &ncommits, npacked,
+		err = load_commit_or_tag(loose_ids, ncommits, npacked,
 		    traversed_ids, id, repo, progress_cb, progress_arg, &rl,
-		    nloose, cancel_cb, cancel_arg);
+		    *nloose, cancel_cb, cancel_arg);
 		if (err)
 			goto done;
 	}
@@ -1200,10 +1238,10 @@ got_repo_purge_unreferenced_loose_objects(struct got_repository *repo,
 	arg.progress_arg = progress_arg;
 	arg.progress_cb = progress_cb;
 	arg.rl = &rl;
-	arg.nloose = nloose;
+	arg.nloose = *nloose;
 	arg.npurged = 0;
 	arg.size_purged = 0;
-	arg.ncommits = ncommits;
+	arg.ncommits = *ncommits;
 	arg.dry_run = dry_run;
 	arg.max_mtime = max_mtime;
 	arg.ignore_mtime = ignore_mtime;
@@ -1214,7 +1252,8 @@ got_repo_purge_unreferenced_loose_objects(struct got_repository *repo,
 
 	/* Produce a final progress report. */
 	if (progress_cb) {
-		err = progress_cb(progress_arg, nloose, ncommits, arg.npurged);
+		err = progress_cb(progress_arg, *nloose, *ncommits, arg.npurged,
+		    -1);
 		if (err)
 			goto done;
 	}
@@ -1225,18 +1264,204 @@ done:
 }
 
 static const struct got_error *
-remove_packidx(int dir_fd, const char *relpath)
+purge_redundant_pack(struct got_repository *repo, const char *packidx_path,
+    int dry_run, int *remove, off_t *size_before, off_t *size_after)
 {
-	const struct got_error *err, *unlock_err;
-	struct got_lockfile *lf;
+	static const char *ext[] = {".idx", ".pack", ".rev", ".bitmap",
+	    ".promisor", ".mtimes"};
+	struct stat sb;
+	char *dot, path[PATH_MAX];
+	size_t i;
 
-	err = got_lockfile_lock(&lf, relpath, dir_fd);
+	if (strlcpy(path, packidx_path, sizeof(path)) >=
+	    sizeof(path))
+		return got_error(GOT_ERR_NO_SPACE);
+
+	/*
+	 * For compatibility with Git, if a matching .keep file exist
+	 * don't delete the packfile.
+	 */
+	dot = strrchr(path, '.');
+	*dot = '\0';
+	if (strlcat(path, ".keep", sizeof(path)) >= sizeof(path))
+		return got_error(GOT_ERR_NO_SPACE);
+	if (faccessat(got_repo_get_fd(repo), path, F_OK, 0) == 0)
+		*remove = 0;
+
+	for (i = 0; i < nitems(ext); ++i) {
+		*dot = '\0';
+
+		if (strlcat(path, ext[i], sizeof(path)) >=
+		    sizeof(path))
+			return got_error(GOT_ERR_NO_SPACE);
+
+		if (fstatat(got_repo_get_fd(repo), path, &sb, 0) ==
+		    -1) {
+			if (errno == ENOENT &&
+			    strcmp(ext[i], ".pack") != 0 &&
+			    strcmp(ext[i], ".idx") != 0)
+				continue;
+			return got_error_from_errno2("fstatat", path);
+		}
+
+		*size_before += sb.st_size;
+		if (!*remove) {
+			*size_after += sb.st_size;
+			continue;
+		}
+
+		if (dry_run)
+			continue;
+
+		if (unlinkat(got_repo_get_fd(repo), path, 0) == -1) {
+			if (errno == ENOENT)
+				continue;
+			return got_error_from_errno2("unlinkat",
+			    path);
+		}
+	}
+
+	return NULL;
+}
+
+static const struct got_error *
+pack_is_redundant(int *redundant, struct got_repository *repo,
+    const char *packidx_path, struct got_object_idset *idset)
+{
+	const struct got_error *err;
+	struct got_packidx *packidx;
+	struct got_packidx_object_id *pid;
+	struct got_object_id id;
+	size_t i, nobjects;
+
+	*redundant = 1;
+
+	err = got_repo_get_packidx(&packidx, packidx_path, repo);
 	if (err)
 		return err;
-	if (unlinkat(dir_fd, relpath, 0) == -1)
-		err = got_error_from_errno("unlinkat");
-	unlock_err = got_lockfile_unlock(lf, dir_fd);
-	return err ? err : unlock_err;
+
+	nobjects = be32toh(packidx->hdr.fanout_table[0xff]);
+	for (i = 0; i < nobjects; ++i) {
+		pid = &packidx->hdr.sorted_ids[i];
+
+		memset(&id, 0, sizeof(id));
+		memcpy(&id.sha1, pid->sha1, sizeof(id.sha1));
+
+		if (got_object_idset_contains(idset, &id))
+			continue;
+
+		*redundant = 0;
+		err = got_object_idset_add(idset, &id, NULL);
+		if (err)
+			return err;
+	}
+
+	return NULL;
+}
+
+struct pack_info {
+	const char	*path;
+	size_t		 nobjects;
+};
+
+static int
+pack_info_cmp(const void *a, const void *b)
+{
+	const struct pack_info	*pa, *pb;
+
+	pa = a;
+	pb = b;
+	if (pa->nobjects == pb->nobjects)
+		return strcmp(pa->path, pb->path);
+	if (pa->nobjects > pb->nobjects)
+		return -1;
+	return 1;
+}
+
+const struct got_error *
+got_repo_purge_redundant_packfiles(struct got_repository *repo,
+    off_t *size_before, off_t *size_after, int dry_run,
+    int nloose, int ncommits, int npurged,
+    got_cleanup_progress_cb progress_cb, void *progress_arg,
+    got_cancel_cb cancel_cb, void *cancel_arg)
+{
+	const struct got_error *err;
+	struct pack_info *pinfo, *sorted = NULL;
+	struct got_packidx *packidx;
+	struct got_object_idset *idset = NULL;
+	struct got_pathlist_entry *pe;
+	size_t i, npacks;
+	int remove, redundant_packs = 0;
+	struct got_ratelimit rl;
+
+	got_ratelimit_init(&rl, 0, 500);
+
+	*size_before = 0;
+	*size_after = 0;
+
+	npacks = 0;
+	TAILQ_FOREACH(pe, &repo->packidx_paths, entry)
+		npacks++;
+
+	if (npacks == 0)
+		return NULL;
+
+	sorted = calloc(npacks, sizeof(*sorted));
+	if (sorted == NULL)
+		return got_error_from_errno("calloc");
+
+	i = 0;
+	TAILQ_FOREACH(pe, &repo->packidx_paths, entry) {
+		err = got_repo_get_packidx(&packidx, pe->path, repo);
+		if (err)
+			goto done;
+
+		pinfo = &sorted[i++];
+		pinfo->path = pe->path;
+		pinfo->nobjects = be32toh(packidx->hdr.fanout_table[0xff]);
+	}
+	qsort(sorted, npacks, sizeof(*sorted), pack_info_cmp);
+
+	idset = got_object_idset_alloc();
+	if (idset == NULL) {
+		err = got_error_from_errno("got_object_idset_alloc");
+		goto done;
+	}
+
+	for (i = 0; i < npacks; ++i) {
+		if (cancel_cb) {
+			err = (*cancel_cb)(cancel_arg);
+			if (err)
+				break;
+		}
+
+		err = pack_is_redundant(&remove, repo, sorted[i].path, idset);
+		if (err)
+			goto done;
+		err = purge_redundant_pack(repo, sorted[i].path, dry_run,
+		    &remove, size_before, size_after);
+		if (err)
+			goto done;
+		if (!remove)
+			continue;
+		err = report_cleanup_progress(progress_cb, progress_arg,
+		    &rl, nloose, ncommits, npurged, ++redundant_packs);
+		if (err)
+			goto done;
+	}
+
+	/* Produce a final progress report. */
+	if (progress_cb) {
+		err = progress_cb(progress_arg, nloose, ncommits, npurged,
+		    redundant_packs);
+		if (err)
+			goto done;
+	}
+ done:
+	free(sorted);
+	if (idset)
+		got_object_idset_free(idset);
+	return err;
 }
 
 const struct got_error *
@@ -1297,9 +1522,10 @@ got_repo_remove_lonely_packidx(struct got_repository *repo, int dry_run,
 		}
 
 		if (!dry_run) {
-			err = remove_packidx(packdir_fd, dent->d_name);
-			if (err)
+			if (unlinkat(packdir_fd, dent->d_name, 0) == -1) {
+				err = got_error_from_errno("unlinkat");
 				goto done;
+			}
 		}
 		if (progress_cb) {
 			char *path;
